@@ -86,6 +86,8 @@ def cli(loop, aiohttp_client):
             'GET', '/resource1/a/b', handler)
         app.router.add_route(
             'GET', '/resource2/a/b/', handler)
+        app.router.add_route(
+            'GET', '/resource2/a/b%2Fc/', handler)
         app.middlewares.extend(extra_middlewares)
         return aiohttp_client(app, server_kwargs={'skip_url_asserts': True})
     return wrapper
@@ -101,7 +103,9 @@ class TestNormalizePathMiddleware:
         ('/resource1?p1=1&p2=2', 200),
         ('/resource1/?p1=1&p2=2', 404),
         ('/resource2?p1=1&p2=2', 200),
-        ('/resource2/?p1=1&p2=2', 200)
+        ('/resource2/?p1=1&p2=2', 200),
+        ('/resource2/a/b%2Fc', 200),
+        ('/resource2/a/b%2Fc/', 200)
     ])
     async def test_add_trailing_when_necessary(
             self, path, status, cli):
@@ -112,6 +116,27 @@ class TestNormalizePathMiddleware:
         resp = await client.get(path)
         assert resp.status == status
 
+    @pytest.mark.parametrize("path, status", [
+        ('/resource1', 200),
+        ('/resource1/', 200),
+        ('/resource2', 404),
+        ('/resource2/', 200),
+        ('/resource1?p1=1&p2=2', 200),
+        ('/resource1/?p1=1&p2=2', 200),
+        ('/resource2?p1=1&p2=2', 404),
+        ('/resource2/?p1=1&p2=2', 200),
+        ('/resource2/a/b%2Fc', 404),
+        ('/resource2/a/b%2Fc/', 200)
+    ])
+    async def test_remove_trailing_when_necessary(self, path, status, cli):
+        extra_middlewares = [
+            web.normalize_path_middleware(
+                append_slash=False, remove_slash=True, merge_slashes=False)]
+        client = await cli(extra_middlewares)
+
+        resp = await client.get(path)
+        assert resp.status == status
+
     @pytest.mark.parametrize("path, status", [
         ('/resource1', 200),
         ('/resource1/', 404),
@@ -120,7 +145,9 @@ class TestNormalizePathMiddleware:
         ('/resource1?p1=1&p2=2', 200),
         ('/resource1/?p1=1&p2=2', 404),
         ('/resource2?p1=1&p2=2', 404),
-        ('/resource2/?p1=1&p2=2', 200)
+        ('/resource2/?p1=1&p2=2', 200),
+        ('/resource2/a/b%2Fc', 404),
+        ('/resource2/a/b%2Fc/', 200)
     ])
     async def test_no_trailing_slash_when_disabled(
             self, path, status, cli):
@@ -194,6 +221,52 @@ class TestNormalizePathMiddleware:
         resp = await client.get(path)
         assert resp.status == status
 
+    @pytest.mark.parametrize("path, status", [
+        ('/resource1/a/b', 200),
+        ('/resource1/a/b/', 200),
+        ('//resource2//a//b', 404),
+        ('//resource2//a//b/', 200),
+        ('///resource1//a//b', 200),
+        ('///resource1//a//b/', 200),
+        ('/////resource1/a///b', 200),
+        ('/////resource1/a///b/', 200),
+        ('/////resource1/a///b///', 200),
+        ('/resource2/a/b', 404),
+        ('//resource2//a//b', 404),
+        ('//resource2//a//b/', 200),
+        ('///resource2//a//b', 404),
+        ('///resource2//a//b/', 200),
+        ('/////resource2/a///b', 404),
+        ('/////resource2/a///b/', 200),
+        ('/resource1/a/b?p=1', 200),
+        ('/resource1/a/b/?p=1', 200),
+        ('//resource2//a//b?p=1', 404),
+        ('//resource2//a//b/?p=1', 200),
+        ('///resource1//a//b?p=1', 200),
+        ('///resource1//a//b/?p=1', 200),
+        ('/////resource1/a///b?p=1', 200),
+        ('/////resource1/a///b/?p=1', 200),
+        ('/resource2/a/b?p=1', 404),
+        ('//resource2//a//b?p=1', 404),
+        ('//resource2//a//b/?p=1', 200),
+        ('///resource2//a//b?p=1', 404),
+        ('///resource2//a//b/?p=1', 200),
+        ('/////resource2/a///b?p=1', 404),
+        ('/////resource2/a///b/?p=1', 200)
+    ])
+    async def test_remove_and_merge_slash(self, path, status, cli):
+        extra_middlewares = [
+            web.normalize_path_middleware(
+                append_slash=False, remove_slash=True)]
+
+        client = await cli(extra_middlewares)
+        resp = await client.get(path)
+        assert resp.status == status
+
+    async def test_cannot_remove_and_add_slash(self):
+        with pytest.raises(AssertionError):
+            web.normalize_path_middleware(append_slash=True, remove_slash=True)
+
 
 async def test_old_style_middleware(loop, aiohttp_client):
     async def handler(request):
